"""
Base database adaptor.
"""

from contextlib import asynccontextmanager
from typing import AsyncContextManager, Awaitable, Callable, Dict, Literal, Optional, Type, Union

from ..enums.databases import Databases
from ..utils.log import Logger
from .connection import AbstractDBConnection
from .oracle_adaptor import OracleDBConnection, OracleSessionPool
from .pool import AbstractSessionPool
from .postgres_adaptor import PostgresDBConnection, PostgresSessionPool


class DBAdaptor:
    """
    Adaptor for database (postgres, oracle)

    Attributes:
        adaptor (AbstractSessionPool): factory for connection to database
        dbType (str): database type
        dsn (str): connection URI in the following format: scheme://user:pass@host:port/database?option=value
        dbtype (str): alias for dbType
        dblogin (str): database login
        dbpassword (str): database password
        adaptorInitKwargs (Dict): some arguments to initialize connection pool
        connectionMaker (Type[Callable[[Logger], Awaitable[AsyncContextManager[AbstractDBConnection]]]]):
            a connection to db maker
        connectionClass (AbstractDBConnection): connection class
        connectionKwargs (Dict): some arguments to pass to the connection maker
        sessionCount (int): session count
    """

    def __init__(
        self,
        dbtype: str,
        dsn: str,
        dblogin: str,
        dbpassword: str,
        sessionCount: int,
        geometryType: Optional[Literal["geometry", "geography"]] = None,
        connectTimeout: int = 5,
    ):
        """
        DB adaptor initialize

        Args:
            dbtype (str): database type ('postgres' or 'oracle')
            dsn (str): connection URI in the following format: scheme://user:pass@host:port/database?option=value
            dblogin (str): database login
            dbpassword (str): database password
            sessionCount (int): session count (for oracle pool)
            geometryType (Optional[str]): geometry type used in the database - "geometry"/"geography" or None
                                                                                                if no geometry is used
            connectTimeout (int): connection timeout, 5 seconds by default
        """
        self.dbType: str = dbtype
        self.dsn: str = dsn
        self.dblogin: str = dblogin
        self.dbpassword: str = dbpassword
        self.adaptor: Union[AbstractSessionPool, None] = None
        self.sessionCount: int = sessionCount
        self.adaptorInitKwargs: Dict = {"sessionCount": sessionCount, "geometryType": geometryType}
        self.connectTimeout = connectTimeout
        self.connectionMaker: Type[Callable[[Logger], Awaitable[AsyncContextManager[AbstractDBConnection]]]] = None
        self.connectionKwargs: Dict = {"sessionCount": sessionCount} if self.dbType != Databases.POSTGRES.value else {}
        self.connectionClass = self.getConnectionClass()

    @property
    def dbtype(self):
        """For backwards compatibility."""
        return self.dbType

    def getConnectionClass(self) -> Type[AbstractDBConnection]:
        """
        Override this to use custom connection classes in subclasses.

        Raises:
            RuntimeError in case dbType is uknown

        Returns:
            Appropriate connection class.
        """
        if self.dbType == Databases.POSTGRES.value:
            return PostgresDBConnection

        if self.dbType == Databases.ORACLE.value:
            return OracleDBConnection

        raise RuntimeError(f"wrong dbType: {self.dbType}")

    async def initiateAdaptor(self):
        """
        Initiate connection pool or engine to database
        """
        kwargs = {
            "dsn": self.dsn,
            "dblogin": self.dblogin,
            "dbpassword": self.dbpassword,
            "connectTimeout": self.connectTimeout,
        }
        if self.dbType == Databases.POSTGRES.value:
            self.adaptor = PostgresSessionPool(**kwargs)
        elif self.dbType == Databases.ORACLE.value:
            self.adaptor = OracleSessionPool(**kwargs)
        else:
            raise RuntimeError(f"wrong dbType: {self.dbType}")
        await self.adaptor.initialize(**self.adaptorInitKwargs)

        self.connectionMaker = self.adaptor.connection

    @asynccontextmanager
    async def connection(self, logger: Logger) -> AsyncContextManager[AbstractDBConnection]:
        """
        Context manager for connection to database. All requests to the database belong to one transaction.

        Args:
            logger: logger
        """
        async with self.connectionMaker(logger) as connection:
            yield self.connectionClass(connection=connection)

    async def close(self) -> None:
        """
        Attempt to gracefully close all connection to db.
        """
        if self.adaptor is not None:
            await self.adaptor.close()
