"""
Base database session pool.
"""

from abc import ABC, abstractmethod
from typing import AsyncGenerator, Dict, Union

from ..utils.log import Logger


class AbstractSessionPool(ABC):
    """
    Wrapper for a db session pool

    Attributes:
        dsn (str): connection URI in the following format: scheme://user:pass@host:port/database?option=value
        dblogin (str): database login
        dbpassword (str): database password
        pool (Pool): connection pool
    """

    def __init__(self, dsn: str, dblogin: str, dbpassword: str):
        self.dsn: str = dsn
        self.dblogin: str = dblogin
        self.dbpassword: str = dbpassword
        self.pool: Union[
            "asyncpg.pool.Pool",
            Dict[str, "oracle_adaptor.OracleThread"],
            None,
        ] = None

    @abstractmethod
    async def initialize(self, sessionCount: int, **kwargs):
        """
        Initialize connection pool
        Args:
            sessionCount: session count
            **kwargs: additional args to initialize connection pool
        """

    @abstractmethod
    async def connection(self, logger: Logger) -> AsyncGenerator["asyncpg.transaction.Transaction", None]:
        """
        Get connection to database
        Args:
            logger: logger
        """

    @abstractmethod
    async def close(self) -> None:
        """
        Attempt to gracefully close all connection to db.
        """

    def printConnectionErrorToLog(self, exception: BaseException, logger: Logger):
        """
        Logs the database connection arguments and the original exception.
        Args:
            exception: database exception
            logger: logger
        Raises:
            database exception
        """
        dbSettings = {
            "dsn": self.dsn,
            "user": self.dblogin,
            "password": "*" * 8,
        }
        logger.error(f"Database error: {exception}, with settings: {dbSettings}")
