"""
Module contains pydantic model for events meta
"""

import re

from pydantic import ConfigDict, model_validator  # pylint: disable-msg=no-name-in-module

from ..db.user_defined_filters import OPERATOR_MAP
from ..errors.pydantic_errors import PydanticError as Error
from ..web.validators import validateMetaFilter
from .bases import JsonBaseModel

META_RX = re.compile(r"^(?P<path>(\w+[.])*([a-zA-Z0-9]+[_]?)+)" rf"(__(?P<operator>({'|'.join(OPERATOR_MAP)})))?$")


class Meta(JsonBaseModel):
    """
    Meta validation class
    """

    model_config = ConfigDict(extra="allow")

    @model_validator(mode="before")
    def mainValidator(cls, values):  # pylint: disable-msg=no-self-argument
        """
        Main validator checks:
            - meta validation
        Args:
            values: loaded values
        Returns:
            loaded values
        """
        if not isinstance(values, dict):
            raise Error.PydanticInvalidDictError()  # pylint: disable-msg=used-before-assignment

        if values:
            try:
                for rawname, rawvalue in values.items():
                    parsed = META_RX.match(rawname)
                    if not parsed:
                        raise Error.PydanticValueError.format("Wrong filter name")()
                    validateMetaFilter(rawvalue, parsed.group("operator") or "eq")
            except Exception as exc:
                raise Error.PydanticValueError.format(
                    f"Validation of filter '{rawname}' has failed. {exc.args[1]}"
                )() from exc

        return values

    def asDict(self) -> dict:
        """
        Overrides super.asDict to prevent using alias for meta filters keys
        """
        return self.model_dump()
