"""Implements base class for RPC server."""

import asyncio
import pickle
from typing import Any, Awaitable, Dict

import msgpack
import ormsgpack
import zmq
import zmq.asyncio

from ..errors.errors import Error
from ..errors.exception import VLException
from ..utils.log import Logger
from .exceptions import RPCException


class Server:
    """Base class for RPC server."""

    def __init__(self, routes: Dict[str, Awaitable], address: str, logger: Logger):
        """
        Args:
            routes: A dictionary with mapping command name -> coroutine
            address: bind address
            logger: A logger instance
        """
        context = zmq.asyncio.Context.instance()
        socket = context.socket(zmq.ROUTER)  # pylint: disable=E1101
        logger.info(f"Binding socket to {address}")
        socket.bind(address)
        self.poll = zmq.asyncio.Poller()
        self.poll.register(socket, zmq.POLLIN)
        self.socket = socket
        self.logger = logger
        self.routes = routes

    def close(self):
        """Close socket."""
        self.socket.close()

    async def serve(self):
        """Main function."""
        while True:
            events = await self.poll.poll()
            requests = []
            for socket, _ in events:
                identity, _, message = await socket.recv_multipart()
                requests.append((identity, message))

            for identity, message in requests:
                asyncio.create_task(self.handleRequest(identity, message))

    async def handleRequest(self, identity: bytes, message: bytes):
        """
        Handle incoming request.

        Args:
            identity: identifier of the client
            message: payload
        """
        try:
            response = await self.handleMessage(message)
            await self.socket.send(identity, zmq.SNDMORE)
            await self.socket.send(b"", zmq.SNDMORE)
            await self.socket.send(response)
        except Exception as exc:  # pylint: disable-msg=broad-except
            # no need to worry about cancelled
            self.logger.exception(exc)
            response = b""
            await self.socket.send(identity, zmq.SNDMORE)
            await self.socket.send(b"", zmq.SNDMORE)
            await self.socket.send(response)

    @staticmethod
    def success(result: Any) -> bytes:
        """Sucess response builder."""
        response = result, None
        return ormsgpack.packb(response)  # pylint: disable-msg=no-member

    @staticmethod
    def error(error: Exception) -> bytes:
        """Error response builder."""
        response = None, pickle.dumps(error)
        return ormsgpack.packb(response)  # pylint: disable-msg=no-member

    async def handleMessage(self, message: bytes) -> bytes:
        """
        Message handler.

        Args:
            messages: bytes
        """

        try:
            requestId, command, args = msgpack.unpackb(message)
        except (ValueError, msgpack.UnpackException) as exc:
            error = f"Couldn't decode passed message: {exc}"
            self.logger.warning(error)
            return self.error(VLException(Error.InvalidCommand))

        if command not in self.routes:
            return self.error(VLException(Error.CommandDoesntExist))

        try:
            handler = self.routes[command]
            if args is not None:
                coro = handler(requestId, *args)
            else:
                coro = handler(requestId)
            result = await coro
        except (VLException, RPCException) as exc:
            return self.error(exc)
        except Exception:  # pylint: disable-msg=broad-except
            self.logger.exception()
            return self.error(VLException(Error.InternalServerError))

        return self.success(result)
